@@ -14,18 +14,23 @@ use serde::{Deserialize, Deserializer, Serialize, Serializer};
 use std::fmt;
 use std::mem;
 use std::ops;
+use std::ptr;
+use std::slice;
 use std::str;
 use std::cmp::Ordering::{self, Equal};
 use std::sync::Mutex;
 use std::sync::atomic::AtomicIsize;
 use std::sync::atomic::Ordering::SeqCst;
 
-use string_cache_shared::{self, UnpackedAtom, Static, Inline, Dynamic, STATIC_ATOM_SET,
-                          ENTRY_ALIGNMENT, copy_memory};
+use shared::{STATIC_TAG, INLINE_TAG, DYNAMIC_TAG, TAG_MASK, MAX_INLINE_LEN, STATIC_SHIFT_BITS,
+             ENTRY_ALIGNMENT, pack_static, StaticAtomSet};
+use self::UnpackedAtom::{Dynamic, Inline, Static};
 
 #[cfg(feature = "log-events")]
 use event::Event;
 
+include!(concat!(env!("OUT_DIR"), "/static_atom_set.rs"));
+
 #[cfg(not(feature = "log-events"))]
 macro_rules! log (($e:expr) => (()));
 
@@ -140,14 +145,16 @@ impl Atom {
     unsafe fn unpack(&self) -> UnpackedAtom {
         UnpackedAtom::from_packed(self.data)
     }
+}
 
+impl<'a> From<&'a str> for Atom {
     #[inline]
-    pub fn from_slice(string_to_add: &str) -> Atom {
+    fn from(string_to_add: &str) -> Atom {
         let unpacked = match STATIC_ATOM_SET.get_index_or_hash(string_to_add) {
             Ok(id) => Static(id as u32),
             Err(hash) => {
                 let len = string_to_add.len();
-                if len <= string_cache_shared::MAX_INLINE_LEN {
+                if len <= MAX_INLINE_LEN {
                     let mut buf: [u8; 7] = [0; 7];
                     copy_memory(string_to_add.as_bytes(), &mut buf);
                     Inline(len as u8, buf)
@@ -161,30 +168,13 @@ impl Atom {
         log!(Event::Intern(data));
         Atom { data: data }
     }
-
-    #[inline]
-    pub fn as_slice<'t>(&'t self) -> &'t str {
-        unsafe {
-            match self.unpack() {
-                Inline(..) => {
-                    let buf = string_cache_shared::inline_orig_bytes(&self.data);
-                    str::from_utf8(buf).unwrap()
-                },
-                Static(idx) => STATIC_ATOM_SET.index(idx).expect("bad static atom"),
-                Dynamic(entry) => {
-                    let entry = entry as *mut StringCacheEntry;
-                    &(*entry).string
-                }
-            }
-        }
-    }
 }
 
 impl Clone for Atom {
     #[inline(always)]
     fn clone(&self) -> Atom {
         unsafe {
-            match string_cache_shared::from_packed_dynamic(self.data) {
+            match from_packed_dynamic(self.data) {
                 Some(entry) => {
                     let entry = entry as *mut StringCacheEntry;
                     (*entry).ref_count.fetch_add(1, SeqCst);
@@ -207,7 +197,7 @@ impl Drop for Atom {
         }
 
         unsafe {
-            match string_cache_shared::from_packed_dynamic(self.data) {
+            match from_packed_dynamic(self.data) {
                 Some(entry) => {
                     let entry = entry as *mut StringCacheEntry;
                     if (*entry).ref_count.fetch_sub(1, SeqCst) == 1 {
@@ -226,7 +216,19 @@ impl ops::Deref for Atom {
 
     #[inline]
     fn deref(&self) -> &str {
-        self.as_slice()
+        unsafe {
+            match self.unpack() {
+                Inline(..) => {
+                    let buf = inline_orig_bytes(&self.data);
+                    str::from_utf8(buf).unwrap()
+                },
+                Static(idx) => STATIC_ATOM_SET.index(idx).expect("bad static atom"),
+                Dynamic(entry) => {
+                    let entry = entry as *mut StringCacheEntry;
+                    &(*entry).string
+                }
+            }
+        }
     }
 }
 
@@ -248,7 +250,7 @@ impl fmt::Debug for Atom {
             }
         };
 
-        write!(f, "Atom('{}' type={})", self.as_slice(), ty_str)
+        write!(f, "Atom('{}' type={})", &*self, ty_str)
     }
 }
 
@@ -258,7 +260,7 @@ impl PartialOrd for Atom {
         if self.data == other.data {
             return Some(Equal);
         }
-        self.as_slice().partial_cmp(other.as_slice())
+        self.as_ref().partial_cmp(other.as_ref())
     }
 }
 
@@ -268,7 +270,7 @@ impl Ord for Atom {
         if self.data == other.data {
             return Equal;
         }
-        self.as_slice().cmp(other.as_slice())
+        self.as_ref().cmp(other.as_ref())
     }
 }
 
@@ -288,7 +290,122 @@ impl Serialize for Atom {
 impl Deserialize for Atom {
     fn deserialize<D>(deserializer: &mut D) -> Result<Atom,D::Error> where D: Deserializer {
         let string: String = try!(Deserialize::deserialize(deserializer));
-        Ok(Atom::from_slice(&*string))
+        Ok(Atom::from(&*string))
+    }
+}
+
+// Atoms use a compact representation which fits this enum in a single u64.
+// Inlining avoids actually constructing the unpacked representation in memory.
+#[allow(missing_copy_implementations)]
+enum UnpackedAtom {
+    /// Pointer to a dynamic table entry.  Must be 16-byte aligned!
+    Dynamic(*mut ()),
+
+    /// Length + bytes of string.
+    Inline(u8, [u8; 7]),
+
+    /// Index in static interning table.
+    Static(u32),
+}
+
+struct RawSlice {
+    data: *const u8,
+    len: usize,
+}
+
+#[cfg(target_endian = "little")]  // Not implemented yet for big-endian
+#[inline(always)]
+unsafe fn inline_atom_slice(x: &u64) -> RawSlice {
+    let x: *const u64 = x;
+    RawSlice {
+        data: (x as *const u8).offset(1),
+        len: 7,
+    }
+}
+
+impl UnpackedAtom {
+    #[inline(always)]
+    unsafe fn pack(self) -> u64 {
+        match self {
+            Static(n) => pack_static(n),
+            Dynamic(p) => {
+                let n = p as u64;
+                debug_assert!(0 == n & TAG_MASK);
+                n
+            }
+            Inline(len, buf) => {
+                debug_assert!((len as usize) <= MAX_INLINE_LEN);
+                let mut data: u64 = (INLINE_TAG as u64) | ((len as u64) << 4);
+                {
+                    let raw_slice = inline_atom_slice(&mut data);
+                    let dest: &mut [u8] = slice::from_raw_parts_mut(
+                        raw_slice.data as *mut u8, raw_slice.len);
+                    copy_memory(&buf[..], dest);
+                }
+                data
+            }
+        }
+    }
+
+    #[inline(always)]
+    unsafe fn from_packed(data: u64) -> UnpackedAtom {
+        debug_assert!(DYNAMIC_TAG == 0); // Dynamic is untagged
+
+        match (data & TAG_MASK) as u8 {
+            DYNAMIC_TAG => Dynamic(data as *mut ()),
+            STATIC_TAG => Static((data >> STATIC_SHIFT_BITS) as u32),
+            INLINE_TAG => {
+                let len = ((data & 0xf0) >> 4) as usize;
+                debug_assert!(len <= MAX_INLINE_LEN);
+                let mut buf: [u8; 7] = [0; 7];
+                let raw_slice = inline_atom_slice(&data);
+                let src: &[u8] = slice::from_raw_parts(raw_slice.data, raw_slice.len);
+                copy_memory(src, &mut buf[..]);
+                Inline(len as u8, buf)
+            },
+            _ => debug_unreachable!(),
+        }
+    }
+}
+
+/// Used for a fast path in Clone and Drop.
+#[inline(always)]
+unsafe fn from_packed_dynamic(data: u64) -> Option<*mut ()> {
+    if (DYNAMIC_TAG as u64) == (data & TAG_MASK) {
+        Some(data as *mut ())
+    } else {
+        None
+    }
+}
+
+/// For as_slice on inline atoms, we need a pointer into the original
+/// string contents.
+///
+/// It's undefined behavior to call this on a non-inline atom!!
+#[inline(always)]
+unsafe fn inline_orig_bytes<'a>(data: &'a u64) -> &'a [u8] {
+    match UnpackedAtom::from_packed(*data) {
+        Inline(len, _) => {
+            let raw_slice = inline_atom_slice(&data);
+            let src: &[u8] = slice::from_raw_parts(raw_slice.data, raw_slice.len);
+            &src[..(len as usize)]
+        }
+        _ => debug_unreachable!(),
+    }
+}
+
+
+/// Copy of std::slice::bytes::copy_memory, which is unstable.
+#[inline]
+fn copy_memory(src: &[u8], dst: &mut [u8]) {
+    let len_src = src.len();
+    assert!(dst.len() >= len_src);
+    // `dst` is unaliasable, so we know statically it doesn't overlap
+    // with `src`.
+    unsafe {
+        ptr::copy_nonoverlapping(src.as_ptr(),
+                                 dst.as_mut_ptr(),
+                                 len_src);
     }
 }
 
@@ -299,32 +416,33 @@ mod bench;
 mod tests {
     use std::mem;
     use std::thread;
-    use super::{Atom, StringCacheEntry};
-    use string_cache_shared::{Static, Inline, Dynamic, ENTRY_ALIGNMENT};
+    use super::{Atom, StringCacheEntry, STATIC_ATOM_SET};
+    use super::UnpackedAtom::{Dynamic, Inline, Static};
+    use shared::ENTRY_ALIGNMENT;
 
     #[test]
     fn test_as_slice() {
-        let s0 = Atom::from_slice("");
-        assert!(s0.as_slice() == "");
+        let s0 = Atom::from("");
+        assert!(s0.as_ref() == "");
 
-        let s1 = Atom::from_slice("class");
-        assert!(s1.as_slice() == "class");
+        let s1 = Atom::from("class");
+        assert!(s1.as_ref() == "class");
 
-        let i0 = Atom::from_slice("blah");
-        assert!(i0.as_slice() == "blah");
+        let i0 = Atom::from("blah");
+        assert!(i0.as_ref() == "blah");
 
-        let s0 = Atom::from_slice("BLAH");
-        assert!(s0.as_slice() == "BLAH");
+        let s0 = Atom::from("BLAH");
+        assert!(s0.as_ref() == "BLAH");
 
-        let d0 = Atom::from_slice("zzzzzzzzzz");
-        assert!(d0.as_slice() == "zzzzzzzzzz");
+        let d0 = Atom::from("zzzzzzzzzz");
+        assert!(d0.as_ref() == "zzzzzzzzzz");
 
-        let d1 = Atom::from_slice("ZZZZZZZZZZ");
-        assert!(d1.as_slice() == "ZZZZZZZZZZ");
+        let d1 = Atom::from("ZZZZZZZZZZ");
+        assert!(d1.as_ref() == "ZZZZZZZZZZ");
     }
 
     macro_rules! unpacks_to (($e:expr, $t:pat) => (
-        match unsafe { Atom::from_slice($e).unpack() } {
+        match unsafe { Atom::from($e).unpack() } {
             $t => (),
             _ => panic!("atom has wrong type"),
         }
@@ -348,17 +466,17 @@ mod tests {
 
     #[test]
     fn test_equality() {
-        let s0 = Atom::from_slice("fn");
-        let s1 = Atom::from_slice("fn");
-        let s2 = Atom::from_slice("loop");
+        let s0 = Atom::from("fn");
+        let s1 = Atom::from("fn");
+        let s2 = Atom::from("loop");
 
-        let i0 = Atom::from_slice("blah");
-        let i1 = Atom::from_slice("blah");
-        let i2 = Atom::from_slice("blah2");
+        let i0 = Atom::from("blah");
+        let i1 = Atom::from("blah");
+        let i2 = Atom::from("blah2");
 
-        let d0 = Atom::from_slice("zzzzzzzz");
-        let d1 = Atom::from_slice("zzzzzzzz");
-        let d2 = Atom::from_slice("zzzzzzzzz");
+        let d0 = Atom::from("zzzzzzzz");
+        let d1 = Atom::from("zzzzzzzz");
+        let d2 = Atom::from("zzzzzzzzz");
 
         assert!(s0 == s1);
         assert!(s0 != s2);
@@ -377,9 +495,9 @@ mod tests {
     #[test]
     fn ord() {
         fn check(x: &str, y: &str) {
-            assert_eq!(x < y, Atom::from_slice(x) < Atom::from_slice(y));
-            assert_eq!(x.cmp(y), Atom::from_slice(x).cmp(&Atom::from_slice(y)));
-            assert_eq!(x.partial_cmp(y), Atom::from_slice(x).partial_cmp(&Atom::from_slice(y)));
+            assert_eq!(x < y, Atom::from(x) < Atom::from(y));
+            assert_eq!(x.cmp(y), Atom::from(x).cmp(&Atom::from(y)));
+            assert_eq!(x.partial_cmp(y), Atom::from(x).partial_cmp(&Atom::from(y)));
         }
 
         check("a", "body");
@@ -395,17 +513,17 @@ mod tests {
 
     #[test]
     fn clone() {
-        let s0 = Atom::from_slice("fn");
+        let s0 = Atom::from("fn");
         let s1 = s0.clone();
-        let s2 = Atom::from_slice("loop");
+        let s2 = Atom::from("loop");
 
-        let i0 = Atom::from_slice("blah");
+        let i0 = Atom::from("blah");
         let i1 = i0.clone();
-        let i2 = Atom::from_slice("blah2");
+        let i2 = Atom::from("blah2");
 
-        let d0 = Atom::from_slice("zzzzzzzz");
+        let d0 = Atom::from("zzzzzzzz");
         let d1 = d0.clone();
-        let d2 = Atom::from_slice("zzzzzzzzz");
+        let d2 = Atom::from("zzzzzzzzz");
 
         assert!(s0 == s1);
         assert!(s0 != s2);
@@ -434,12 +552,11 @@ mod tests {
     #[test]
     fn repr() {
         fn check(s: &str, data: u64) {
-            assert_eq_fmt!("0x{:016X}", Atom::from_slice(s).data, data);
+            assert_eq_fmt!("0x{:016X}", Atom::from(s).data, data);
         }
 
         fn check_static(s: &str, x: Atom) {
-            use string_cache_shared::STATIC_ATOM_SET;
-            assert_eq_fmt!("0x{:016X}", x.data, Atom::from_slice(s).data);
+            assert_eq_fmt!("0x{:016X}", x.data, Atom::from(s).data);
             assert_eq!(0x2, x.data & 0xFFFF_FFFF);
             // The index is unspecified by phf.
             assert!((x.data >> 32) <= STATIC_ATOM_SET.iter().len() as u64);
@@ -450,9 +567,9 @@ mod tests {
         // static atom table, the tag values, etc.
 
         // Static atoms
-        check_static("a",       atom!(a));
-        check_static("address", atom!(address));
-        check_static("area",    atom!(area));
+        check_static("a",       atom!("a"));
+        check_static("address", atom!("address"));
+        check_static("area",    atom!("area"));
 
         // Inline atoms
         check("e",       0x0000_0000_0000_6511);
@@ -460,7 +577,7 @@ mod tests {
         check("xyzzy01", 0x3130_797A_7A79_7871);
 
         // Dynamic atoms. This is a pointer so we can't verify every bit.
-        assert_eq!(0x00, Atom::from_slice("a dynamic string").data & 0xf);
+        assert_eq!(0x00, Atom::from("a dynamic string").data & 0xf);
     }
 
     #[test]
@@ -475,36 +592,35 @@ mod tests {
     fn test_threads() {
         for _ in 0_u32..100 {
             thread::spawn(move || {
-                let _ = Atom::from_slice("a dynamic string");
-                let _ = Atom::from_slice("another string");
+                let _ = Atom::from("a dynamic string");
+                let _ = Atom::from("another string");
             });
         }
     }
 
     #[test]
     fn atom_macro() {
-        assert_eq!(atom!(body), Atom::from_slice("body"));
-        assert_eq!(atom!("body"), Atom::from_slice("body"));
-        assert_eq!(atom!("font-weight"), Atom::from_slice("font-weight"));
+        assert_eq!(atom!("body"), Atom::from("body"));
+        assert_eq!(atom!("font-weight"), Atom::from("font-weight"));
     }
 
     #[test]
     fn match_atom() {
-        assert_eq!(2, match Atom::from_slice("head") {
-            atom!(br) => 1,
-            atom!(html) | atom!(head) => 2,
+        assert_eq!(2, match Atom::from("head") {
+            atom!("br") => 1,
+            atom!("html") | atom!("head") => 2,
             _ => 3,
         });
 
-        assert_eq!(3, match Atom::from_slice("body") {
-            atom!(br) => 1,
-            atom!(html) | atom!(head) => 2,
+        assert_eq!(3, match Atom::from("body") {
+            atom!("br") => 1,
+            atom!("html") | atom!("head") => 2,
             _ => 3,
         });
 
-        assert_eq!(3, match Atom::from_slice("zzzzzz") {
-            atom!(br) => 1,
-            atom!(html) | atom!(head) => 2,
+        assert_eq!(3, match Atom::from("zzzzzz") {
+            atom!("br") => 1,
+            atom!("html") | atom!("head") => 2,
             _ => 3,
         });
     }
@@ -512,14 +628,14 @@ mod tests {
     #[test]
     fn ensure_deref() {
         // Ensure we can Deref to a &str
-        let atom = Atom::from_slice("foobar");
+        let atom = Atom::from("foobar");
         let _: &str = &atom;
     }
 
     #[test]
     fn ensure_as_ref() {
         // Ensure we can as_ref to a &str
-        let atom = Atom::from_slice("foobar");
+        let atom = Atom::from("foobar");
         let _: &str = atom.as_ref();
     }
 
@@ -529,7 +645,7 @@ mod tests {
     #[cfg(feature = "unstable")]
     #[test]
     fn atom_drop_is_idempotent() {
-        use string_cache_shared::from_packed_dynamic;
+        use super::from_packed_dynamic;
         unsafe {
             assert_eq!(from_packed_dynamic(mem::POST_DROP_U64), None);
         }
