@@ -214,6 +214,7 @@ pub static ATOMS: &'static [&'static str] = &[
     "bbox",
     "bdi",
     "bdo",
+    "beforeunload",
     "begin",
     "bevelled",
     "bgcolor",
@@ -696,6 +697,7 @@ pub static ATOMS: &'static [&'static str] = &[
     "moment",
     "momentabout",
     "movablelimits",
+    "mouseover",
     "mover",
     "mozbrowser",
     "mpadded",
@@ -738,6 +740,7 @@ pub static ATOMS: &'static [&'static str] = &[
     "notin",
     "notprsubset",
     "notsubset",
+    "novalidate",
     "nowrap",
     "number",
     "numoctaves",
